c ParaDyn - Parallel DYNAMO - molecular dynamics with EAM potentials
c 
c Authored by Steve Plimpton
c   (505) 845-7873, sjplimp@cs.sandia.gov
c   Dept 9221, MS 1111, Sandia National Labs, Albuquerque, NM  87185-1111
c
c Based on the serial DYNAMO code authored by
c   Stephen Foiles (foiles@ca.sandia.gov), Sandia National Labs, Livermore, CA
c   Murray Daw (daw@hubcap.clemson.edu), Clemson University
c
c See the README file for more information
c
      
      subroutine setup_parallel
#include "paradyn.h"
#include "mpif.h"

c divide up atoms among procs

      nstart = nint(float(node)/nprocs*natoms) + 1
      nend = nint(float(node+1)/nprocs*natoms)
      nlocal = nend - nstart + 1

c error check

c map procs into numrow by numcol 2-d array
c setup row and column communicators
c do-loop inside pmeth=2 (force-decomposition) chooses largest factors
c  of nprocs to form squarish numrow by numcol
      
      if (pmeth.eq.1) then
        numcol = 1
        numrow = nprocs
      else
        numcolmax = nint(sqrt(float(nprocs)))
        do i = 1,numcolmax
          if (mod(nprocs,i).eq.0) numcol = i
        enddo
        numrow = nprocs/numcol
      endif

      if (node.eq.0) then
        write (6,*) 'Partitioning of procs: Nrow =',
     $       numrow,' Ncol =',numcol
        if (logflag.eq.1)
     $       write (1,*) 'Partitioning of procs: Nrow =',
     $       numrow,' Ncol =',numcol
      endif

      irow = node/numcol
      call mpi_comm_split(mpi_comm_world,irow,0,row_comm,ierror)

      icol = mod(node,numcol)
      call mpi_comm_split(mpi_comm_world,icol,0,col_comm,ierror)

c nrow = # of atoms owned by all procs in my row
c ncol = # of atoms owned by all procs in my col

      call mpi_allreduce(nlocal,nrow,1,mpi_integer,mpi_sum,
     $     row_comm,ierror)
      call mpi_allreduce(nlocal,ncol,1,mpi_integer,mpi_sum,
     $     col_comm,ierror)

c allocate memory (DYNAMIC mode only)

      call setup_memory

c error checks on array sizes

      call mpi_allreduce(nlocal,nmax,1,mpi_integer,mpi_max,
     $     mpi_comm_world,ierror)
      if (nmax.gt.maxlocal)
     $     call error('Too many atoms/processor - boost maxlocal')

      call mpi_allreduce(nrow,nmax,1,mpi_integer,mpi_max,
     $     mpi_comm_world,ierror)
      if (nmax.gt.maxrow)
     $     call error('Too many atoms/row - boost maxrow')

      call mpi_allreduce(ncol,nmax,1,mpi_integer,mpi_max,
     $     mpi_comm_world,ierror)
      if (nmax.gt.maxcol)
     $     call error('Too many atoms/column - boost maxcol')

c create plans for gather/scatter in rows and columns of procs

      call plan_gather_scatter(1,nlocal,row_comm,plan_row)
      call plan_gather_scatter(1,3*nlocal,row_comm,plan_row3)
#ifdef STRESS
      call plan_gather_scatter(1,6*nlocal,row_comm,plan_row6)
#endif

      call plan_gather_scatter(1,nlocal,col_comm,plan_col)
      call plan_gather_scatter(1,3*nlocal,col_comm,plan_col3)
#ifdef STRESS
      call plan_gather_scatter(1,6*nlocal,col_comm,plan_col6)
#endif

      return
      end


c if DYNAMIC flag is set, allocate all arrays

      subroutine setup_memory
#include "paradyn.h"
#include "mpif.h"

#ifdef DYNAMIC

c initialize all dynamic array pointers to 0

      data x_ptr,xtmp_ptr,xrow_ptr,xcol_ptr,xold_ptr /5*0/
      data v_ptr /0/
      data tag_ptr,tagtmp_ptr,tagrow_ptr,tagcol_ptr /4*0/
      data itype_ptr,itypetmp_ptr,ityperow_ptr,itypecol_ptr /4*0/
      data f_ptr,ftmp_ptr,frow_ptr,fcol_ptr /4*0/
      data e_ptr,etmp_ptr,erow_ptr,ecol_ptr /4*0/
      data rho_ptr,rhotmp_ptr,rhorow_ptr,rhocol_ptr /4*0/
      data fp_ptr,fprow_ptr,fpcol_ptr /3*0/
      data nsmall_pt_ptr,jneigh_ptr,dneigh_ptr,rneigh_ptr /4*0/
      data nlarge_pt_ptr,nlarge_lst_ptr /2*0/
      data bin_ptr /0/
      data parity_ptr,u1_ptr,u2_ptr,u3_ptr,u4_ptr,u5_ptr,
     $     y_ptr,acc_ptr /8*0/
      data tmptag_ptr,tmpseed_ptr,fixtag_ptr /3*0/
#ifdef STRESS
      data strs_ptr,strstmp_ptr,strsrow_ptr,strscol_ptr /4*0/
#endif

c free memory from any previous run

      call memfree(x_ptr)
      call memfree(xtmp_ptr)
      call memfree(xrow_ptr)
      call memfree(xcol_ptr)
      call memfree(xold_ptr)

      call memfree(v_ptr)

      call memfree(tag_ptr)
      call memfree(tagtmp_ptr)
      call memfree(tagrow_ptr)
      call memfree(tagcol_ptr)

      call memfree(itype_ptr)
      call memfree(itypetmp_ptr)
      call memfree(ityperow_ptr)
      call memfree(itypecol_ptr)

      call memfree(f_ptr)
      call memfree(ftmp_ptr)
      call memfree(frow_ptr)
      call memfree(fcol_ptr)

      call memfree(e_ptr)
      call memfree(etmp_ptr)
      call memfree(erow_ptr)
      call memfree(ecol_ptr)

      call memfree(rho_ptr)
      call memfree(rhotmp_ptr)
      call memfree(rhorow_ptr)
      call memfree(rhocol_ptr)

      call memfree(fp_ptr)
      call memfree(fprow_ptr)
      call memfree(fpcol_ptr)

      call memfree(nsmall_pt_ptr)
      call memfree(jneigh_ptr)
      call memfree(dneigh_ptr)
      call memfree(rneigh_ptr)

      call memfree(nlarge_pt_ptr)
      call memfree(nlarge_lst_ptr)

      call memfree(bin_ptr)

      call memfree(parity_ptr)
      call memfree(u1_ptr)
      call memfree(u2_ptr)
      call memfree(u3_ptr)
      call memfree(u4_ptr)
      call memfree(u5_ptr)
      call memfree(y_ptr)
      call memfree(acc_ptr)

      call memfree(tmptag_ptr)
      call memfree(tmpseed_ptr)
      call memfree(fixtag_ptr)

#ifdef STRESS
      call memfree(strs_ptr)
      call memfree(strstmp_ptr)
      call memfree(strsrow_ptr)
      call memfree(strscol_ptr)
#endif

      call memset(0)

c compute current bounds

      call mpi_allreduce(nlocal,maxlocal,1,mpi_integer,mpi_max,
     $     mpi_comm_world,ierror)
      call mpi_allreduce(nrow,maxrow,1,mpi_integer,mpi_max,
     $     mpi_comm_world,ierror)
      call mpi_allreduce(ncol,maxcol,1,mpi_integer,mpi_max,
     $     mpi_comm_world,ierror)

      density = 4.0D0 / latconst**3
      volume = 4.0D0/3.0D0 * 3.1415926D0 * rcutmax**3
      maxnsmall = volume * density * extra_neigh
      volume = 4.0D0/3.0D0 * 3.1415926D0 * (rcutmax+dradn)**3
      maxnlarge = volume * density * extra_neigh

      if (newton.eq.2) then
        maxnsmall = maxnsmall/2 + 1
        maxnlarge = maxnlarge/2 + 1
      endif

c allocate all arrays

      call memalloc(x_ptr,3*maxlocal,1)
      call memalloc(xtmp_ptr,3*maxlocal,1)
      call memalloc(xrow_ptr,3*maxrow,1)
      call memalloc(xcol_ptr,3*maxcol,1)
      call memalloc(xold_ptr,3*maxlocal,1)

      call memalloc(v_ptr,3*maxlocal,1)

      call memalloc(tag_ptr,maxlocal,0)
      call memalloc(tagtmp_ptr,maxlocal,0)
      call memalloc(tagrow_ptr,maxrow,0)
      call memalloc(tagcol_ptr,maxcol,0)

      call memalloc(itype_ptr,maxlocal,0)
      call memalloc(itypetmp_ptr,maxlocal,0)
      call memalloc(ityperow_ptr,maxrow,0)
      call memalloc(itypecol_ptr,maxcol,0)

      call memalloc(f_ptr,3*(maxlocal+1),1)
      call memalloc(ftmp_ptr,3*maxlocal,1)
      call memalloc(frow_ptr,3*maxrow,1)
      call memalloc(fcol_ptr,3*maxcol,1)

      call memalloc(e_ptr,maxlocal+1,1)
      call memalloc(etmp_ptr,maxlocal,1)
      call memalloc(erow_ptr,maxrow,1)
      call memalloc(ecol_ptr,maxcol,1)

      call memalloc(rho_ptr,maxlocal,1)
      call memalloc(rhotmp_ptr,maxlocal,1)
      call memalloc(rhorow_ptr,maxrow,1)
      call memalloc(rhocol_ptr,maxcol,1)

      call memalloc(fp_ptr,maxlocal,1)
      call memalloc(fprow_ptr,maxrow,1)
      call memalloc(fpcol_ptr,maxcol,1)

      call memalloc(nsmall_pt_ptr,maxrow+1,0)
      call memalloc(jneigh_ptr,maxnsmall*maxlocal+1000,0)
      call memalloc(dneigh_ptr,3*(maxnsmall*maxlocal+1000),1)
      call memalloc(rneigh_ptr,maxnsmall*maxlocal+1000,1)

      call memalloc(nlarge_pt_ptr,maxrow+1,0)
      call memalloc(nlarge_lst_ptr,maxnlarge*maxlocal+1000,0)

      call memalloc(bin_ptr,maxcol,0)

      call memalloc(parity_ptr,3*maxlocal,0)
      call memalloc(u1_ptr,3*(maxlocal+1),1)
      call memalloc(u2_ptr,3*(maxlocal+1),1)
      call memalloc(u3_ptr,3*(maxlocal+1),1)
      call memalloc(u4_ptr,3*(maxlocal+1),1)
      call memalloc(u5_ptr,3*(maxlocal+1),1)
      call memalloc(y_ptr,6*(maxlocal+1),1)
      call memalloc(acc_ptr,3*(maxlocal+1),1)

      call memalloc(tmptag_ptr,maxlocal,0)
      call memalloc(tmpseed_ptr,maxlocal,0)
      call memalloc(fixtag_ptr,maxlocal,0)

#ifdef STRESS
      call memalloc(strs_ptr,6*maxlocal,1)
      call memalloc(strstmp_ptr,6*maxlocal,1)
      call memalloc(strsrow_ptr,6*maxrow,1)
      call memalloc(strscol_ptr,6*maxcol,1)
#endif

#endif

c set bottom value of neighbor pointers

      nsmall_pt(0) = 0
      nlarge_pt(0) = 0

      return
      end
